@@ -326,12 +326,15 @@ class Panoply:
                 # some op commands do not properly wrap data in CDATA tags, so let's try to
                 # replace some chars manually and re-parse...
                 x = self.xapi.xml_document
-                cx = x.replace('&', '&amp;')
+
+                # show config audit version will return an embedded xml document which breaks everything
+                # just remove the invalid embedded xml tags if present
+                cx = x.replace('&', '&amp;').replace('<?xml version="1.0"?>', '').replace('</xml>', '')
 
                 try:
                     # make sure our little 'fix' didn't completely ruin the xml structure and return what we got
                     s = ElementTree.fromstring(cx)
-                    return ElementTree.tostring(s)
+                    return ElementTree.tostring(s).decode(encoding='UTF-8')
 
                 except Exception as e:
                     logger.error(e)
@@ -617,6 +620,11 @@ class Panoply:
             facts['dns-primary'] = '1.1.1.1'
             facts['dns-secondary'] = '1.0.0.1'
 
+        try:
+            facts['panorama-server'] = results['system']['panorama']['local-panorama']['panorama-server']
+        except KeyError:
+            facts['panorama-server'] = None
+
         return facts
 
     def load_baseline(self) -> bool:
@@ -1065,7 +1073,10 @@ class Panoply:
             cmd = 'show config candidate'
 
         else:
-            cmd = 'show config running'
+            if self.facts.get('panorama-server', None) is not None:
+                cmd = 'show config merged'
+            else:
+                cmd = 'show config running'
 
         try:
 
@@ -1542,6 +1553,14 @@ class Panoply:
 
     @staticmethod
     def __check_children_are_list(c: list) -> bool:
+        """
+        Check if this list possibly contains identical children. I.E. a list of members or simple entries.
+        If this is the case, then we'll use xmldiff to find diffs in there to keep this logic simple. Otherwise,
+        we'll continue to iterate through the document
+
+        :param c: list of elements
+        :return: True if all elements have the same tag name and have no children
+        """
         # check if children are a list of items by identical tag names
 
         if len(c) <= 1:
@@ -1552,6 +1571,10 @@ class Panoply:
 
         for child in c:
 
+            # if any child itself has children, then this is not a list of identical items. Return False here
+            if len(child):
+                return False
+
             if found_tag_name == '':
                 found_tag_name = child.tag
                 continue
