@@ -1,4 +1,5 @@
 import logging
+import re
 import shlex
 import time
 import traceback
@@ -25,6 +26,7 @@ class DockerSnippet(Snippet):
 
     # optional parameters that may be set in the snippet metadata
     optional_metadata = {
+        'tag': 'latest',
         'volumes': dict(),
         'async': True
     }
@@ -47,7 +49,7 @@ class DockerSnippet(Snippet):
 
         else:
             self.image = image
-            self.tag = 'latest'
+            self.tag = self.metadata.get('tag', 'latest')
 
         self.working_dir = self.metadata.get('working_dir', '/app')
 
@@ -92,7 +94,6 @@ class DockerSnippet(Snippet):
         """
         try:
 
-            output = ''
             logger.info(f'Pulling image: {self.image} with tag: {self.tag}')
             self.client.images.pull(self.image, self.tag)
 
@@ -102,6 +103,7 @@ class DockerSnippet(Snippet):
             logger.info('Creating container...')
             return_data = self.client.containers.run(image, self.metadata['cmd'], volumes=vols, stderr=True,
                                                      detach=self.detach, working_dir=self.working_dir,
+                                                     user=self.metadata.get('user', 'root'),
                                                      auto_remove=self.auto_remove, environment=context)
 
             if self.detach:
@@ -115,9 +117,9 @@ class DockerSnippet(Snippet):
                 # return_data will be the bytes returned from the command
                 if type(return_data) is bytes:
                     return_str = return_data.decode('UTF-8')
-                    return return_str, 'success'
+                    return return_str, self.__get_container_status()
                 else:
-                    return return_data, 'success'
+                    return return_data, self.__get_container_status()
 
         except ImageNotFound:
             logger.error(traceback.format_exc())
@@ -165,14 +167,33 @@ class DockerSnippet(Snippet):
 
             else:
                 logger.info(container.status)
-                return return_str, 'success'
+                return return_str, self.__get_container_status()
 
         except APIError as ae:
             raise SkilletLoaderException(f'Could not get logs for {self.name}: {ae}')
 
+    def __get_container_status(self) -> str:
+        """
+        Check for ExitCode State on the container and return the exit code if found
+
+        :return: success or failure. Return code of 0 == success all else == failure
+        """
+
+        container = self.get_container()
+        if container.status != 'running':
+            rc = container.attrs['State']['ExitCode']
+            if rc == 0:
+                return 'success'
+            else:
+                return 'failure'
+
+        else:
+            return 'success'
+
     def cleanup(self) -> None:
         """
         Clean up action is the docker container was started with 'async', no-op is async is False
+
         :return: None
         """
         if not self.detach:
@@ -199,7 +220,9 @@ class DockerSnippet(Snippet):
             return str(return_data)
 
     # Fix for GL #77 - escape cmd variables before passing into render_metadata
-    def render_metadata(self, context: dict) -> dict:
+    # removed due to quote adding extra quotes around items with spaces. The recommendation is to add / use
+    # the quote filter in your cmd strings when you know you may get special chars
+    def render_metadata_deprecated(self, context: dict) -> dict:
         """
         render_metadata will ensure all metadata vars that contain jinja2 variables will be rendered properly.
         i.e. cmd: ansible-playbook -e 'somevar={{ some_value }}' will be converted to 'somevar='value_from_context'
@@ -213,9 +236,13 @@ class DockerSnippet(Snippet):
         # only get the variables that are used in the 'cmd'
         affected_vars = self.get_variables_from_template(self.metadata['cmd'])
 
+        all_non_safe = re.compile(r'[^\w|\s]')
+
         # iterate over each var, get it's value from the context, quote it, and set back into the context
         for v in affected_vars:
             if v in context:
-                context[v] = shlex.quote(context[v])
+                # only quote things that have special chars, otherwise leave it alone
+                if all_non_safe.search(context[v]):
+                    context[v] = shlex.quote(context[v])
 
         return super().render_metadata(context)
