@@ -1,5 +1,7 @@
 from typing import Tuple
 
+from jinja2.exceptions import TemplateError
+
 from .base import Snippet
 
 
@@ -12,6 +14,11 @@ class TemplateSnippet(Snippet):
     output_type = 'text'
 
     template_metadata = {'element'}
+    optional_metadata = {
+        'file': '',
+        'element': '',
+        'template_title': ''
+    }
 
     def __init__(self, template_str, metadata):
         self.template_str = template_str
@@ -19,10 +26,18 @@ class TemplateSnippet(Snippet):
         super().__init__(metadata)
 
     def execute(self, context: dict) -> Tuple[str, str]:
-        return self.render(self.template_str, context), 'success'
+        try:
+            return self.render(self.template_str, context), 'success'
+
+        except TemplateError as te:
+            return str(te), 'failure'
 
     def template(self, context) -> str:
-        return self.execute(context)[0]
+        try:
+            o = self.execute(context)
+            return o[0]
+        except TemplateError as te:
+            return str(te)
 
 
 class SimpleTemplateSnippet(TemplateSnippet):
